import os
import pytest
from fhir.resources import FHIRAbstractModel

from fhir_kindling import FhirServer, FHIRQuery
from dotenv import load_dotenv, find_dotenv
from fhir.resources.organization import Organization
from fhir.resources.address import Address
from fhir.resources.bundle import Bundle, BundleEntry, BundleEntryRequest
from fhir.resources.patient import Patient
from fhir_kindling.generators import PatientGenerator


@pytest.fixture
def api_url():
    """
    Base api url and env vars
    """
    load_dotenv(find_dotenv())

    return os.getenv("FHIR_API_URL", "http://test.fhir.org/r4")


@pytest.fixture
def oidc_server(api_url):
    server = FhirServer(
        api_address=api_url,
        client_id=os.getenv("CLIENT_ID"),
        client_secret=os.getenv("CLIENT_SECRET"),
        oidc_provider_url=os.getenv("OIDC_PROVIDER_URL")
    )
    return server


@pytest.fixture
def fhir_server(api_url):
    server = FhirServer(
        api_address=api_url,
        client_id=os.getenv("CLIENT_ID"),
        client_secret=os.getenv("CLIENT_SECRET"),
        oidc_provider_url=os.getenv("OIDC_PROVIDER_URL")
    )
    return server


@pytest.fixture
def org_bundle(api_url):
    bundle = Bundle.construct()
    bundle.type = "transaction"

    entries = []

    for i in range(10):
        org = Organization.construct()
        org.name = f"Test Create Org No. {i}"
        address = Address.construct()
        address.country = "Germany"
        org.address = [address]

        entry = BundleEntry.construct()

        entry.request = BundleEntryRequest(
            **{
                "method": "POST",
                "url": org.get_resource_type()
            }
        )

        entry.resource = org.dict()
        entries.append(entry)

    bundle.entry = entries
    return bundle


def test_server_api_url_validation(api_url):
    server = FhirServer(api_address=api_url)

    assert server
    malformed_url_1 = "hello-world"

    with pytest.raises(ValueError):
        server = FhirServer(api_address=malformed_url_1)

    malformed_url_2 = "htp://hapi.fhir.org/baseR4"
    with pytest.raises(ValueError):
        server = FhirServer(api_address=malformed_url_2)

    server = FhirServer(api_address="http://blaze:8080/fhir")
    assert server

    server = FhirServer(api_address="http://bla_ze:8080/fhir")
    assert server

    server = FhirServer(api_address="http://bla-ze:8080/fhir")
    assert server

    with pytest.raises(ValueError):
        server = FhirServer(api_address="http://bla ze:8080/fhir")

    with pytest.raises(ValueError):
        server = FhirServer(api_address="http://bla_ze:8080/ fhir")

    with pytest.raises(ValueError):
        server = FhirServer(api_address="http://bla_ze:8x80/fhir")


def test_server_capabilities(fhir_server):
    capabilities = fhir_server.capabilities
    assert capabilities


def test_upload_single_resource(fhir_server: FhirServer):
    from fhir.resources.organization import Organization
    from fhir.resources.address import Address

    org = Organization.construct()
    org.name = "Test Create Org"
    address = Address.construct()
    address.country = "Germany"
    org.address = [address]

    Organization.element_properties()

    response = fhir_server.add(org)

    print(response.resource)

    assert response


def test_server_add_all(fhir_server: FhirServer):
    generator = PatientGenerator(n=100)
    patients = generator.generate()
    add_response = fhir_server.add_all(patients)

    assert add_response


def test_query_all(fhir_server: FhirServer):
    response = fhir_server.query(Organization, output_format="dict").all()

    assert response["entry"]


def test_query_with_string_resource(fhir_server: FhirServer):
    auth = fhir_server.auth
    query = FHIRQuery(fhir_server.api_address, "Patient", auth=auth)

    response = query.all()

    assert isinstance(query.resource, FHIRAbstractModel)

    assert response


def test_query_with_limit(fhir_server: FhirServer):
    response = fhir_server.query(Patient, output_format="dict").limit(2)

    assert response["entry"]

    assert len(response["entry"]) == 2


def test_query_raw_string(fhir_server: FhirServer):
    query_string = "/Patient?"
    query = fhir_server.raw_query(query_string=query_string, output_format="dict")

    assert isinstance(query, FHIRQuery)

    assert query.resource.get_resource_type() == "Patient"

    assert query.all()["entry"]


def test_add_bundle(fhir_server: FhirServer, org_bundle):
    response = fhir_server.add_bundle(org_bundle)
    assert response
    print(response)



import os
from pathlib import Path
from typing import List, Union

from fhir.resources.bundle import Bundle
from fhir.resources.organization import Organization
from fhir.resources.reference import Reference
from fhir.resources.identifier import Identifier

from fhir_kindling.generators import FhirResourceGenerator, PatientResourceGenerator, MolecularSequenceGenerator, \
    PatientGenerator
from .upload import upload_bundle, upload_resource
from dotenv import load_dotenv, find_dotenv


# TODO clean up authentication flow

def generate_data_set(name: str, generators: List[FhirResourceGenerator] = None, bundle: Union[Bundle, str] = None,
                      fhir_api_url: str = None,
                      username: str = None,
                      password: str = None,
                      token: str = None,
                      fhir_server_type: str = "hapi",
                      out_dir: Union[str, Path] = None, filename: str = None):
    if not (generators or bundle):
        raise ValueError("Either bundle or generators need to be given to generate a data set.")

    # create organisation to group generated patients under
    organisation = create_organisation_for_data_set(name)

    # register it with the server and get the reference
    r, organisation_reference = upload_resource(organisation,
                                                fhir_api_url=fhir_api_url,
                                                username=username,
                                                password=password,
                                                token=token,
                                                fhir_server_type=fhir_server_type,
                                                reference=True
                                                )

    if generators:
        exit_code = _make_data_set_with_generators(generators=generators, organisation_reference=organisation_reference,
                                                   fhir_api_url=fhir_api_url, username=username, password=password,
                                                   token=token, fhir_server_type=fhir_server_type)

        return exit_code

    elif bundle:
        # TODO upload bundle as data set
        pass

    return 0


def _make_data_set_with_generators(generators: List[FhirResourceGenerator],
                                   organisation_reference: Reference,
                                   fhir_api_url: str = None,
                                   username: str = None,
                                   password: str = None,
                                   token: str = None,
                                   fhir_server_type: str = "hapi"
                                   ):
    num_patients = max([gen.num_patients for gen in generators])
    print(f"Generating {num_patients} Patients for the defined resources.")
    # Generate patients and upload bundle to get server assigned ids
    patient_generator = PatientGenerator(n=num_patients, organisation=organisation_reference)
    patients = patient_generator.generate()
    response, patient_references = upload_bundle(bundle=patient_generator.make_bundle(), fhir_api_url=fhir_api_url,
                                                 username=username, password=password, token=token,
                                                 fhir_server_type=fhir_server_type,
                                                 references=True)

    for gen in generators:
        resources = gen.generate(patient_references=patient_references)
        response = upload_bundle(gen.make_bundle(), fhir_api_url=fhir_api_url, username=username, password=password,
                                 token=token, fhir_server_type=fhir_server_type)

    return 0


def generate_resources(generator: Union[FhirResourceGenerator, PatientResourceGenerator],
                       patient_references: List[str] = None, fhir_api_url: str = None,
                       out_dir: Union[str, Path] = None, filename: str = None):
    if isinstance(generator, FhirResourceGenerator):
        resources = generator.generate(out_dir=out_dir, filename=filename)
        return resources, generator.make_bundle()
    # Generate patients for resources and initially register
    elif isinstance(generator, PatientResourceGenerator):
        if not patient_references:
            patients_bundle = generator.generate_patients(bundle=True)
            response, patient_references = upload_bundle(patients_bundle, references=True, fhir_api_url=fhir_api_url)

        resources, bundle = generator.generate(patient_references, out_dir, filename)
        return resources, bundle


def create_organisation_for_data_set(name: str) -> Organization:
    org = Organization(
        **{
            "active": True,
            "name": name
        }
    )
    return org


if __name__ == '__main__':
    load_dotenv(find_dotenv())
    # sequence_file = "../examples/hiv_sequences/sequences_2.txt"
    # mol_seq_generator = MolecularSequenceGenerator(sequence_file=sequence_file)
    # patient_ms_generator = PatientResourceGenerator(resource_generator=mol_seq_generator)
    # resources, bundle = generate_resources(patient_ms_generator, fhir_api_url=os.getenv("FHIR_API_URL"), out_dir=".",
    #                                        filename="sequence_bundle.json")
    #
    # upload_bundle(bundle=bundle, fhir_api_url=os.getenv("FHIR_API_URL"))
    sequence_file_5 = "../../examples/hiv_sequences/sequences_5.txt"
    sequence_file_4 = "../../examples/hiv_sequences/sequences_4.txt"
    sequence_file_3 = "../../examples/hiv_sequences/sequences_3.txt"
    ms_generator = MolecularSequenceGenerator(sequence_file=[sequence_file_3, sequence_file_4])

    dataset = generate_data_set(name="HIV_DEMO", generators=[ms_generator],
                                fhir_api_url="https://ibm-fhir.personalhealthtrain.de/fhir-server/api/v4",
                                fhir_server_type="ibm")



import importlib
import inspect
import os
from pathlib import Path

# TODO re-export all fhir resources under a single module and apply some transformation on them


if __name__ == '__main__':
    resources = importlib.import_module("fhir.resources")
    print(resources.__file__)

    resource_module_path = Path(resources.__file__).parent
    for mod in list(os.listdir(resource_module_path)):
        print(mod)
        if mod.split(".")[-1] == "py":
            module = importlib.import_module(f"fhir.resources.{mod.split('.')[0]}")
            members = inspect.getmembers(module)




import os
from typing import List, Union

import requests
from fhir.resources.bundle import BundleEntry, BundleEntryRequest

from fhir_kindling.fhir_server.auth import generate_auth
from fhir.resources import FHIRAbstractModel
from .query_functions import query_server
from .upload import generate_fhir_headers
from dotenv import load_dotenv, find_dotenv


def delete_resources(query: str = None, resource: Union[str, FHIRAbstractModel] = None, cascade: bool = False,
                     fhir_api_url: str = None, username: str = None, password: str = None, token: str = None,
                     fhir_server_type: str = "hapi"):
    pass


def delete_resource_by_type(resource: Union[str, FHIRAbstractModel],
                            fhir_api_url: str = None, username: str = None, password: str = None, token: str = None,
                            fhir_server_type: str = "hapi"
                            ):
    headers = generate_fhir_headers(fhir_server_type)
    auth = generate_auth(username, password, token)
    response, references = query_server(resource=resource, fhir_server_url=fhir_api_url, username=username, password=password,
                                        token=token, references=True, fhir_server_type=fhir_server_type)

    transaction = _make_delete_transaction(references)

    delete_response = requests.post(fhir_api_url, json=transaction, headers=headers, auth=auth)

    return delete_response


def _make_delete_transaction(urls: List[str]) -> dict:
    bundle_requests = []
    for url in urls:
        entry_request = BundleEntryRequest(
            **{
                "method": "DELETE",
                "url": url
            }
        )

        entry = BundleEntry(
            request=entry_request
        )
        bundle_requests.append(entry)

    bundle = {
        "type": "transaction",
        "resourceType": "Bundle",
        "entry": [entry.dict() for entry in bundle_requests]
    }
    return bundle


if __name__ == '__main__':
    load_dotenv(find_dotenv())
    delete_resource_by_type(resource="MolecularSequence", fhir_api_url=os.getenv("IBM_API_URL"),
                            username=os.getenv("FHIR_USER"), password=os.getenv("FHIR_PW"), fhir_server_type="ibm")



#!/usr/bin/env python

"""Tests for `fhir_kindling` package."""

import pytest

from click.testing import CliRunner

import fhir_kindling


@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string


# def test_command_line_interface():
#     """Test the CLI."""
#     runner = CliRunner()
#     result = runner.invoke(fhir_kindling.cli)
#     assert result.exit_code == 0
#     assert 'fhir_kindling.cli.main' in result.output
#     help_result = runner.invoke(fhir_kindling.main, ['--help'])
#     assert help_result.exit_code == 0
#     assert '--help  Show this message and exit.' in help_result.output



from typing import List

from fhir.resources.bundle import Bundle
from fhir.resources.domainresource import DomainResource

from fhir_kindling.resource_generator import FhirResourceGenerator

class FHIRMediaGenerator(FhirResourceGenerator):
    def __init__(self, n: int, resources: List[DomainResource] = None, resource_type: DomainResource = None,
                 fhir_server: str = None, fhir_user: str = None, fhir_pw: str = None, fhir_token: str = None,
                 fhir_server_type: str = None):
        super().__init__(n, resources, resource_type, fhir_server, fhir_user, fhir_pw, fhir_token, fhir_server_type)

    def generate(self, upload: bool = False, out_dir: str = None):
        super().generate(upload, out_dir)

    def make_bundle(self) -> Bundle:
        return super().make_bundle()



import json
from pathlib import Path
from typing import Union, List, Tuple
import pandas as pd

import requests
from requests.auth import AuthBase
from fhir.resources.domainresource import DomainResource
from fhir.resources.fhirtypes import DomainResourceType
from fhir.resources import FHIRAbstractModel

from fhir_kindling.fhir_server.auth import generate_auth, load_environment_auth_vars
from fhir_kindling.serde import flatten_bundle


# todo clean up authentication flow

def query_server(query_string: str = None,
                 resource: Union[FHIRAbstractModel, DomainResourceType, str] = None,
                 limit: int = 1000,
                 out_path: Union[str, Path] = None,
                 out_format: str = "json",
                 references: bool = False,
                 fhir_server_url: str = None, username: str = None, password: str = None, token: str = None,
                 fhir_server_type: str = None) -> Union[dict, Tuple[dict, list], pd.DataFrame]:
    """
    Execute a query against a server, either query all instances of a fhir resource or execute a given query string.
    Optionally store the results in a file either in csv format or as a raw fhir bundle json

    Args:
        query_string: Prebuilt query string to execute against the server's REST api
        resource: either string identifier of a resource
        limit: maximum number of resources to return, 0 returns all resources
        out_path: path under which to store the results
        out_format: format in which to store the results either csv or json
        references: whether to return a list of references to the resources
        fhir_server_url: url of the fhir server api endpoint
        username: username for basic auth
        password: password for basic auth
        token: token for bearer token auth
        fhir_server_type: type of the fhir server one of [hapi, blaze, ibm]

    Returns:
        dictionary containing the json response from the server, if references is True returns a tuple
        (bundle, references)
    """
    # Attempt to load environment authentication variables if nothing is given
    if not (username and password) and not token:
        username, password, token = load_environment_auth_vars()

    if query_string and resource:
        raise ValueError("Conflicting query information, both resource and query string are given.")

    # Create authentication and headers for the requests
    auth = generate_auth(username, password, token)
    headers = {"Content-Type": "application/fhir+json"}

    # If a resource type or identifier string is given query this resource or query by given url string
    if resource:
        response = query_resource(resource, fhir_server_url, auth, headers, limit=limit)

    elif query_string:
        response = query_with_string(query_string, fhir_server_url, auth, headers, limit=limit)

    else:
        raise ValueError("No query information given.")

    if out_path:
        if out_format == "json":
            with open(out_path, "w") as out_file:
                json.dump(response, out_file)
        elif out_format == "csv":
            df = flatten_bundle(bundle_json=response)
            df.to_csv(out_path)

        else:
            raise NotImplementedError("Only csv and json currently supported.")

    if out_format == "df":
        df = flatten_bundle(bundle_json=response)
        return df

    if response and references:
        references = _extract_references_from_query_response(entries=response.get("entry"),
                                                             fhir_server_type=fhir_server_type)
        return response, references

    return response


def query_resource(resource: Union[DomainResource, DomainResourceType, str], fhir_server_url: str, auth: AuthBase,
                   headers: dict, fhir_server_type: str = None, count: int = 2000, limit: int = 0):
    """
    Query all instances up to limit from the server.

    Args:
        resource: Either string identifier of a resource or a FHIR resource type or instances
        fhir_server_url: api endpoint to execute the query against
        auth: authentication object for the request
        headers: headers to add to the request
        fhir_server_type: type of the fhir server
        count: maximum resources per page returned by the fhir server
        limit: maximum number of resources to return, all if 0

    Returns:
        bundle containing all instances of the resource on the server

    """
    if isinstance(resource, str):
        url = fhir_server_url + "/" + resource + "?"
    else:
        url = fhir_server_url + "/" + resource.get_resource_type() + "?"

    url += f"_count={limit if limit != 0 and limit < count else count}"
    response = _execute_query(url, auth, headers, limit)

    return response


def query_with_string(query_string: str, fhir_server_url: str, auth: AuthBase, headers: dict, count: int = 2000,
                      limit: int = None) -> dict:
    """
    Execute the given query string using the given fhir api endpoint and return the results

    Args:
        query_string: prebuilt query string to append to the api url and query
        fhir_server_url: api endpoint of the fhir server
        auth: authentication object for the fhir server
        headers: headers to add to the request
        count: maximum number of resources per page
        limit: maximum number of resources to return, all if 0

    Returns:
        bundle containing resources matching the query
    """
    # stitch together the query string with the api url
    if fhir_server_url[-1] == "/" and query_string[0] == "/":
        url = fhir_server_url[:-1] + query_string
    elif fhir_server_url[-1] != "/" and query_string[0] != "/":
        url = fhir_server_url + "/" + query_string
    else:
        url = fhir_server_url + query_string

    if "_count" not in query_string:
        # Set the count to the limit if limit is given
        if limit:
            limit_count = limit if count > limit != 0 else count
            url += f"&_count={limit_count}"
            response = _execute_query(url, auth, headers, limit)
        else:
            response = _execute_query(url, auth, headers, count)

    return response


def _execute_query(url: str, auth: AuthBase = None, headers: dict = None, limit: int = None) -> dict:
    r = requests.get(url=url, auth=auth, headers=headers)
    r.raise_for_status()

    response = r.json()
    # Query additional pages contained in the response and append all returned lists into a list of entries
    response = _resolve_response_pagination(response, auth, headers, limit)

    return response


def _resolve_response_pagination(response: dict, auth: AuthBase, headers: dict, limit: int = None) -> dict:
    entries = []
    entries.extend(response["entry"])

    if limit:
        if len(entries) >= limit:
            return response

    while response.get("link", None):

        if limit != 0 and len(entries) >= limit:
            print("Limit reached stopping pagination resolve")
            break

        next_page = next((link for link in response["link"] if link.get("relation", None) == "next"), None)
        if next_page:
            response = requests.get(next_page["url"], auth=auth, headers=headers).json()
            entries.extend(response["entry"])
        else:
            break

    response["entry"] = entries[:limit] if limit else entries

    return response


def _extract_references_from_query_response(entries: List[dict], fhir_server_type: str):
    references = []
    if fhir_server_type == "hapi":
        for entry in entries:
            subject = entry.get("subject")
            if subject:
                references.append(subject["reference"])
            else:
                references.append(None)

    elif fhir_server_type in ["blaze", "ibm"]:
        for entry in entries:
            resource_reference = "/".join(entry["fullUrl"].split("/")[-2:])
            references.append(resource_reference)
    else:
        raise ValueError(f"Unsupported FHIR server: {fhir_server_type}")

    return references



import pathlib
from typing import Union, Tuple
import requests
from fhir.resources import FHIRAbstractModel
from fhir.resources.bundle import Bundle
from fhir.resources.reference import Reference
from pathlib import Path
from requests.auth import AuthBase

from fhir_kindling.fhir_server.auth import generate_auth
from fhir_kindling.serde import load_bundle
from dotenv import load_dotenv, find_dotenv
import os


def upload_bundle(bundle: Union[Bundle, Path, str],
                  fhir_api_url: str,
                  validate: bool = True,
                  username: str = None,
                  password: str = None,
                  token: str = None,
                  fhir_server_type: str = "hapi",
                  references: bool = False) -> Union[dict, Tuple[dict, list]]:
    """Upload a bundle to a FHIR server.

    Args:
      bundle: Either a bundle object, the path to json file containing a bundle or to a directory containing multiple bundle json files
      fhir_api_url: base url of the FHIR servers api
      validate: flag indicating whether to validate bundles loaded from file
      username: username for basic auth authentication
      password: password for basic auth
      token: token for bearer token authentication
      fhir_server_type: FHIR server implementation defaults to HAPI
      references: indicates whether to return references to the generated resources

    Returns:
      The fhir server's response(s) and references to the uploaded resources if the flag is set.

    """
    auth = generate_auth(username=username, password=password, token=token, load_env=True)

    if isinstance(bundle, str) or isinstance(bundle, Path):

        if pathlib.Path.is_dir(Path(bundle)):
            p = pathlib.Path(bundle).glob("**/*")
            files = [x for x in p if x.is_file()]
            response_dict = {}
            for file in files:
                bundle_file = load_bundle(file)
                bundle_response = _upload_bundle(bundle_file, api_url=fhir_api_url, auth=auth,
                                                 fhir_server_type=fhir_server_type)
                response_dict[str(file)] = bundle_response

            response = response_dict
        else:
            loaded_bundle = load_bundle(bundle)
            response = _upload_bundle(loaded_bundle, api_url=fhir_api_url, auth=auth, fhir_server_type=fhir_server_type)

    else:

        response = _upload_bundle(bundle, api_url=fhir_api_url, auth=auth, fhir_server_type=fhir_server_type)
    if references:
        resource_references = _get_references_from_bundle_response(response)
        return response, resource_references
    else:
        return response


def upload_resource(resource: FHIRAbstractModel,
                    fhir_api_url: str,
                    username: str = None,
                    password: str = None,
                    token: str = None,
                    fhir_server_type: str = "hapi",
                    reference: bool = True):
    """
    Upload a single resource to the server

    Args:
        resource: Resource to uplaod
        fhir_api_url: base url of the fhir rest api to use
        username: username for basic auth
        password: password for basic auth
        token: token to use for bearer auth
        fhir_server_type: type of the fhir server one of [ibm, hapi, blaze]
        reference: whether to return the resource reference separately

    Returns:
        the response to from the server or if reference is set (response, reference)
    """
    auth = generate_auth(username=username, password=password, token=token, load_env=True)
    url = fhir_api_url + f"/{resource.resource_type}"

    r = requests.post(url=url, json=resource.dict(), headers=generate_fhir_headers(fhir_server_type), auth=auth)

    if fhir_server_type == "ibm":
        if reference:
            location = r.headers["location"].split("/")
            print(location[-4:-2])
            resource_reference = Reference(
                **{"reference": f"{location[-4]}/{location[-3]}",
                   "type": location[-4]}
            )

            return r.headers, resource_reference
        return r.headers
    else:

        r.raise_for_status()
        response = r.json()
        if reference:
            resource_reference = Reference(
                **{"reference": f"{response['resourceType']}/{response['id']}",
                   "type": response['resourceType']}
            )
            return response, resource_reference

        return response


def _get_references_from_bundle_response(response):
    references = []
    for entry in response["entry"]:
        location = entry["response"]["location"]
        reference = "/".join(location.split("/")[:2])

        references.append(reference)
    return references


def _upload_bundle(bundle: Union[Bundle, dict], api_url: str, auth: AuthBase, fhir_server_type: str):
    headers = generate_fhir_headers(fhir_server_type)

    if isinstance(bundle, Bundle):

        r = requests.post(api_url, auth=auth, data=bundle.json(), headers=headers)
    else:
        r = requests.post(api_url, auth=auth, data=bundle, headers=headers)
    r.raise_for_status()

    return r.json()


def generate_fhir_headers(fhir_server_type: str):
    headers = {}
    if fhir_server_type == "blaze":
        headers["Content-Type"] = "application/fhir+json"

    else:
        headers["Content-Type"] = "application/fhir+json"

    return headers




"""Console script for fhir_kindling."""
import sys
import click
import yaml
from fhir_kindling.generators import PatientGenerator
from .upload import upload_bundle
from .query_functions import query_server as execute_query
from fhir_kindling.fhir_server.auth import load_environment_auth_vars
from pathlib import Path
from dotenv import load_dotenv, find_dotenv
import click_spinner


@click.group()
def main():
    """Command line interface for generating synthetic FHIR resources and uploading them to a FHIR server."""
    load_dotenv(find_dotenv())


@main.command()
@click.option("-f", "--file", default=None, help="Path to a .yml file defining the generation specs.")
@click.option("-n", "--n-patients", default=None, help="How many patients to generate", type=int)
@click.option("-a", "--age-range", default=None, help="Space separated min/max age of patients.",
              type=click.Tuple([int, int]))
@click.option("-o", "--output", default=None, help="Path where the generated resource bundle should be stored.")
@click.option("--upload", is_flag=True)
@click.option("--url", default=None, help="url of the FHIR api endpoint to upload the bundle to.")
@click.option("--username", default=None, help="username for basic auth")
@click.option("--password", default=None, help="password for basic auth")
@click.option("--token", default=None, help="token for bearer token auth")
def generate(file, n_patients, age_range, output, url, upload, username, password, token):
    """Generate FHIR resource bundles and synthetic data sets"""
    if file:
        click.echo(f"Generating FHIR resources defined in:\n{file}")
        with open(file, "r") as f:
            resource_specs = yaml.safe_load(f)
            click.echo(resource_specs)
    else:
        if not n_patients:
            n_patients = click.prompt("Enter the number of patients you want to create", default=100)

        # Prompt for age range if not given
        if not age_range:
            min_age = click.prompt("Enter the min age patients", default=18, type=int)
            max_age = click.prompt("Enter the max age patients", default=101, type=int)
            age_range = (min_age, max_age)

        # Generate the patients
        patient_generator = PatientGenerator(n_patients, age_range=age_range)
        patients = patient_generator.generate()

        # TODO if FHIR server is given upload the resources to get server generated ids
        if click.confirm("Generate additional resources for patients?"):
            patient_resource = click.prompt("Select resource:", type=click.Choice(["Observation", "Condition"]))
        else:
            pass

    if upload:
        if not url:
            click.prompt("Enter your FHIR server`s API url")

    if not output:
        if click.confirm("No storage location given. Exit without saving generated resources to disk?"):
            return 0
        else:
            output = click.prompt("Enter the path or filename under which the bundle should be stored",
                                  default="bundle.json")
            if Path(output).exists():
                overwrite = click.confirm(f"File already exists. Overwrite {output}?")
                if not overwrite:
                    output = click.prompt("Enter new filepath")
            with open(output, "w") as output_file:
                bundle = patient_generator.make_bundle()
                output_file.write(bundle.json(indent=2))
    click.echo(f"Storing resources in {output}")

    return 0


@main.command()
@click.argument("bundle")
@click.option("--url", help="url of the FHIR api endpoint to upload the bundle to.")
@click.option("-u", "--username", default=None, help="Password to get when authenticating against basic auth.")
@click.option("-p", "--password", default=None, help="Username to use when authenticating with basic auth.")
@click.option("-t", "--token", default=None, help="Token to use with bearer token auth.")
@click.option("-s", "--summary", is_flag=True, help="Print summary information about the upload.")
def upload(bundle, url, username, password, token, summary):
    """Upload a bundle to a fhir server"""

    # Get the url
    if not url:
        url = click.prompt("Enter the API endpoint of the fhir server you want to upload to.")

    if not username and not password and not token:
        click.echo("Attempting to load auth from environment.")
        username, password, token = load_environment_auth_vars()
    if not username and not password and not token:
        click.confirm("No auth information found. Upload without authentication?")

    if (username or password) and token:
        raise ValueError("Only one of basic or token auth can be used.")

    if username and not password:
        password = click.prompt(f"Enter your password ({username}):", hide_input=True)

    click.echo("Uploading bundle...", nl=False)
    with click_spinner.spinner():

        upload_bundle(bundle, fhir_api_url=url, username=username, password=password, token=token)
    click.echo("Resources uploaded successfully")

    return 0


@main.command()
@click.option("-q", "--query", default=None, help="FHIR API query string, appended to api url.")
@click.option("-r", "--resource", default=None, help="Identifier of the resource to query e.g. Patient")
@click.option("--url", help="url of the FHIR api endpoint to query against.")
@click.option("-f", "--file", default=None, help="File in which to save the query results")
@click.option("-o", "--output_format", default="json", help="Format in which to store the results.")
@click.option("-u", "--username", default=None, help="Username to get when authenticating with basic auth.")
@click.option("-p", "--password", default=None, help="Password to use when authenticating with basic auth.")
@click.option("--token", default=None, help="Token to use with bearer token auth.")
def query(query, resource, url, file, output_format, username, password, token):
    """Query resources from a fhir server"""

    if not (username and password) or token:
        click.echo("Attempting to find authentication in environment variables...", err=True, nl=False)
        username, password, token = load_environment_auth_vars()

        if (username and password) or token:
            click.echo("Success")
        else:
            click.echo()
            click.confirm("Continue without authentication?")

    click.echo("Executing query...", nl=False)
    with click_spinner.spinner():
        if query:
            # todo fix fhir server type and solve this in a nicer a way
            response = execute_query(query_string=query, out_path=file, out_format=output_format,
                                     fhir_server_type="hapi",
                                     username=username, password=password, fhir_server_url=url, token=token)
        if resource:
            response = execute_query(resource=resource, out_path=file, out_format=output_format,
                                     fhir_server_type="hapi",
                                     username=username, password=password, fhir_server_url=url, token=token)
    if not file:
        click.echo()
        click.echo_via_pager(response)
    else:
        click.echo()
        click.echo(f"Stored results at {file}")

if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



import json

import pytest
import requests
from fhir.resources.patient import Patient
from requests.auth import HTTPBasicAuth
import os
from unittest import mock

# from fhir_kindling.fhir_query import query_server, query_resource, query_with_string
# from fhir.resources.procedure import Procedure
# from fhir_kindling.upload import generate_fhir_headers
# from fhir_kindling.fhir_server.auth import generate_auth
# import pandas as pd
#
#
# @pytest.fixture
# def api_url():
#     """
#     Base api url
#
#     """
#     return "http://hapi.fhir.org/baseR4"
#
#
# @pytest.fixture
# def bogus_auth():
#     """
#     Lower level functions expect auth information to be given. The public test server does not require but it needs to
#     be present.
#     Returns:
#
#     """
#
#     return HTTPBasicAuth(username="bogus", password="auth")
#
#
# @pytest.fixture
# def fhir_headers():
#     return generate_fhir_headers(fhir_server_type="hapi")
#
#
# def test_api_connection(api_url):
#     url = api_url + "/metadata"
#     r = requests.get(url)
#
#     r.raise_for_status()
#     assert r.text
#
#
# def test_query_resource(api_url, bogus_auth, fhir_headers):
#     limit = 100
#     response = query_resource("Patient", fhir_server_url=api_url, auth=bogus_auth, headers=fhir_headers, limit=limit)
#     assert response
#     assert len(response["entry"]) == limit
#
#     token_auth = generate_auth(token="test_token")
#     response2 = query_resource(Patient, fhir_server_url=api_url, auth=token_auth, headers=fhir_headers, limit=limit)
#     assert response == response2
#
#     # test paginated responses
#     response3 = query_resource("Patient", fhir_server_url=api_url, auth=bogus_auth, headers=fhir_headers, limit=2500)
#     assert len(response3["entry"]) == 2500
#
#
# def test_query_with_string(api_url, bogus_auth, fhir_headers):
#     # simple query for patients by age
#     query_string_1 = "Patient?birthdate=gt1990"
#
#     response_1 = query_with_string(query_string_1, api_url, bogus_auth, fhir_headers, limit=100)
#
#     assert response_1
#     assert len(response_1["entry"]) == 100
#
#     # test that / mismatch is resolved
#     query_string_2 = "/Observation?patient.birthdate=gt1990"
#     response_2 = query_with_string(query_string_2, api_url + "/", bogus_auth, fhir_headers, limit=100)
#
#     assert response_2
#
#
# def test_query(api_url, fhir_headers):
#     query_string = "/Observation?patient.birthdate=gt1990"
#
#     response = query_server(query_string=query_string, limit=100, username="test", password="password",
#                             fhir_server_url=api_url, fhir_server_type="hapi", references=False)
#
#     assert response
#
#     response = query_server(resource="Observation", limit=100, username="test", password="password",
#                             fhir_server_url=api_url, fhir_server_type="hapi", references=False)
#
#     assert response
#
#     with pytest.raises(ValueError):
#         response = query_server(limit=100, username="test", password="password", fhir_server_url=api_url,
#                                 fhir_server_type="hapi", references=False)
#
#     with pytest.raises(ValueError):
#         response = query_server(query_string=query_string, limit=100,
#                                 fhir_server_url=api_url, fhir_server_type="hapi", references=False)
#     with pytest.raises(ValueError):
#         response = query_server(query_string=query_string, limit=100, username="test", password="pw", token="hello",
#                                 fhir_server_url=api_url, fhir_server_type="hapi", references=False)
#
#
# @mock.patch.dict(os.environ, {"FHIR_USER": "test", "FHIR_PW": "password"})
# def test_query_with_env(api_url, fhir_headers):
#     query_string = "/Observation?patient.birthdate=gt1990"
#     response = query_server(query_string=query_string, limit=100, fhir_server_url=api_url, fhir_server_type="hapi",
#                             references=False)
#     assert response
#
#
# def test_query_output(api_url, tmp_path):
#     query_string = "/Observation?patient.birthdate=gt1990"
#
#     json_path = tmp_path / "query_results.json"
#     csv_path = tmp_path / "query_results.csv"
#
#     response = query_server(query_string=query_string, limit=100, fhir_server_url=api_url, fhir_server_type="hapi",
#                             token="hello",
#                             out_path=json_path, out_format="json", references=False)
#
#     loaded_response = json.loads(json_path.read_bytes())
#
#     assert loaded_response
#     assert len(loaded_response["entry"]) == len(response["entry"])
#
#     # test serialization to csv
#     response = query_server(query_string=query_string, limit=100, fhir_server_url=api_url, fhir_server_type="hapi",
#                             token="hello",
#                             out_path=csv_path, out_format="csv", references=False)
#
#     df = pd.read_csv(csv_path)
#
#     assert len(df) == len(response["entry"])
#
#     with pytest.raises(NotImplementedError):
#         response = query_server(query_string=query_string, limit=100, fhir_server_url=api_url, fhir_server_type="hapi",
#                                 token="hello",
#                                 out_path=csv_path, out_format="hdf", references=False)
#
#
# def test_query_references(api_url):
#     query_string = "/Observation?patient.birthdate=gt1990"
#
#     response, references = query_server(query_string=query_string, limit=100, fhir_server_url=api_url, fhir_server_type="hapi",
#                                         token="hello", references=True)
#
#     assert len(response["entry"]) == len(references)



import functools
import inspect
import json
import os
from typing import Union, List

from dotenv import load_dotenv, find_dotenv
from fhir.resources.resource import Resource
from fhir.resources.bundle import Bundle
import fhir.resources
import requests
import requests.auth
import xml.etree.ElementTree as ET


class FHIRQuery:

    def __init__(self,
                 base_url: str,
                 resource: Union[Resource, fhir.resources.FHIRAbstractModel, str] = None,
                 auth: requests.auth.AuthBase = None,
                 session: requests.Session = None,
                 output_format: str = "json"):

        self.base_url = base_url

        # Set up the requests session with auth and headers
        self.auth = auth
        if session:
            self.session = session
        else:
            self._setup_session()

        # initialize the resource
        if isinstance(resource, str):
            self.resource = fhir.resources.get_fhir_model_class(resource)
        else:
            self.resource = resource
        self.resource = self.resource.construct()

        self.output_format = output_format
        self._query_string = None
        self.conditions: List[str] = []
        self._limit = None
        self._query_response: Union[Bundle, str] = None

    def where(self, filter_dict: dict = None):
        # todo evaluate arbitrary number of expressions based on fields of the resource and query values
        self._parse_filter_dict(filter_dict)
        return self

    def _setup_session(self):
        self.session = requests.Session()
        self.session.auth = self.auth
        self.session.headers.update({"Content-Type": "application/fhir+json"})

    def include(self):
        pass

    def has(self):
        pass

    def all(self):
        self._limit = None
        self._execute_query()
        return self._serialize_output()

    def limit(self, n: int):
        self._limit = n
        self._execute_query()
        return self._serialize_output()

    def first(self):
        self._limit = 1
        self._execute_query()
        return self._serialize_output()

    def set_query_string(self, raw_query_string: str):
        self._query_string = self.base_url + raw_query_string

    @property
    def query_url(self):
        if not self._query_string:
            self._make_query_string()
        return self._query_string

    def _execute_query(self):
        r = self.session.get(self.query_url)
        r.raise_for_status()
        if self.output_format == "xml":
            self._query_response = r.content
        else:
            response = self._resolve_response_pagination(r)
            self._query_response = Bundle(**response)

    def _resolve_response_pagination(self, server_response: requests.Response):
        # todo outsource into search response class
        response = server_response.json()
        link = response.get("link", None)
        if not link:
            return response
        else:
            print("Resolving response pagination")
            entries = []
            entries.extend(response["entry"])

            if self._limit:
                if len(entries) >= self._limit:
                    response["entry"] = response["entry"][:self._limit]
                    return response

            while response.get("link", None):

                if self._limit and len(entries) >= self._limit:
                    print("Limit reached stopping pagination resolve")
                    break

                next_page = next((link for link in response["link"] if link.get("relation", None) == "next"), None)
                if next_page:
                    response = self.session.get(next_page["url"]).json()
                    entries.extend(response["entry"])
                else:
                    break

            response["entry"] = entries[:self._limit] if self._limit else entries
            # remove next linked from resolved pagination
            return response

    def _make_query_string(self):
        query_string = self.base_url + "/" + self.resource.get_resource_type() + "?"

        if self.conditions:
            # todo
            pass
        # todo implement include and has
        if self._limit:
            query_string += f"_count={self._limit}"
        else:
            query_string += f"_count=5000"

        # todo improve xml support with full parser
        if self.output_format == "xml":
            query_string += f"&_format=xml"
        else:
            query_string += f"&_format=json"

        self._query_string = query_string

    def _parse_filter_dict(self, filter_dict: dict):
        pass

    def _serialize_output(self):
        # TODO improve the super super basic xml support
        if isinstance(self._query_response, bytes):
            self._query_response = self._query_response.decode("utf-8")
        if isinstance(self._query_response, str):
            if self.output_format == "xml":
                return self._query_response
            elif self.output_format == "json":
                return Bundle.construct(**json.loads(self._query_response)).json()
            elif self.output_format == "dict":
                return Bundle.construct(**json.loads(self._query_response)).dict()
            elif self.output_format == "model":
                return Bundle.construct(**json.loads(self._query_response))

        elif isinstance(self._query_response, Bundle):
            if self.output_format == "json":
                return self._query_response.json(indent=2)
            elif self.output_format == "dict":
                return self._query_response.dict()
            elif self.output_format == "model":
                return self._query_response

        else:
            raise ValueError("Unable to serialize query response")



from .resource_generator import FhirResourceGenerator
from .patient import PatientGenerator, PatientResourceGenerator
from .observation import ObservationGenerator
from .molecular_sequence import MolecularSequenceGenerator



import math
import os
from dotenv import load_dotenv, find_dotenv
from typing import Union, Tuple, List
import pendulum
import pandas as pd
import random
from tqdm import tqdm
from pathlib import Path
from pendulum import DateTime

from fhir.resources.reference import Reference
from fhir.resources.patient import Patient
from fhir.resources.humanname import HumanName

from fhir_kindling.generators import FhirResourceGenerator
from fhir_kindling.fhir_server import FhirServer


class PatientGenerator(FhirResourceGenerator):
    def __init__(self,
                 n: int,
                 age_range: Union[Tuple[DateTime, DateTime], Tuple[int, int]] = None,
                 gender_distribution: Tuple[float, float, float, float] = None,
                 organisation: Reference = None,
                 generate_ids: bool = False):
        super().__init__(n, resource_type=Patient)
        self.age_range = age_range
        self.gender_distribution = gender_distribution
        self.birthdate_range = None
        self.organisation = organisation
        self.generate_ids = generate_ids

    def _generate(self, display: bool = False):
        patients = []
        names = self._generate_patient_names(self.n)
        for i in tqdm(range(self.n), desc=f"Generating {self.n} patients", disable=display):
            patient = self._generate_patient_data(name=names[i])
            patients.append(patient)
        return patients

    def _generate_patient_data(self, name: Tuple[str, str]) -> Patient:
        gender = random.choices(
            ["male", "female", "other", "unknown"],
            weights=self.gender_distribution if self.gender_distribution else [0.45, 0.45, 0.1, 0.0], k=1)[0]

        name = HumanName(**{"family": name[1], "given": [name[0]]})

        birthdate = self._generate_birthdate()
        patient_dict = {
            "gender": gender,
            "name": [name],
            "birthDate": birthdate
        }
        if self.organisation:
            patient_dict["managingOrganization"] = self.organisation

        return Patient(**patient_dict)

    @staticmethod
    def _generate_patient_names(n: int):

        p = Path(__file__).parent.joinpath("data").joinpath("first_names.txt")
        with open(p, "rb") as fnf:
            first_name_list = [fn.decode().strip().capitalize() for fn in fnf.readlines()]
        p = Path(__file__).parent.joinpath("data").joinpath("last_names.txt")
        with open(p, "rb") as lnf:
            last_name_list = [ln.decode().strip().capitalize() for ln in lnf.readlines()]

        first_names = random.choices(first_name_list, k=n)
        last_names = random.choices(last_name_list, k=n)
        names = list(zip(first_names, last_names))

        return names

    def _generate_birthdate(self):
        if not self.birthdate_range:
            if self.age_range:
                if isinstance(self.age_range[0], int):
                    # generate age range from 18-101 years old
                    now = pendulum.now()
                    youngest = pd.to_datetime((now - pendulum.duration(years=self.age_range[0])).to_date_string())
                    oldest = pd.to_datetime((now - pendulum.duration(years=self.age_range[1])).to_date_string())

                elif isinstance(self.age_range[0], DateTime):
                    youngest = self.age_range[0].to_date_string()
                    oldest = self.age_range[1].to_date_string()
                else:
                    raise ValueError(f"Unsupported type ({type(self.age_range[0])}) for generating patient ages."
                                     f"Only integers and datetime are supported.")
            else:
                # generate age range from 18-101 years old
                now = pendulum.now()
                youngest = pd.to_datetime((now - pendulum.duration(years=18)).to_date_string())
                oldest = pd.to_datetime((now - pendulum.duration(years=101)).to_date_string())

            self.birthdate_range = pd.date_range(oldest, youngest, freq="D").strftime('%Y-%m-%d').tolist()

        birthdate = random.choice(self.birthdate_range)
        return birthdate


class PatientResourceGenerator:

    def __init__(self, resource_generator: FhirResourceGenerator = None,
                 patients: Union[List[Patient], List[str], bool] = None,
                 n_per_patient: int = 1):
        self.patients = patients
        self.n_per_patients = n_per_patient
        self.resource_generator = resource_generator
        self.n = self.resource_generator.n
        self.resources = None

    def generate(self, patients=None, out_dir: str = None, filename: str = None, generate_ids: bool = False):
        if patients is None and self.patients is None:
            raise ValueError("No patients given to generate Resources for.")
        else:
            self.patients = patients
            # TODO use self patients and serialize newly generated patients to list
            self.resources = self.resource_generator.generate(generate_ids=generate_ids)
            self.update_with_patient_ids()
            if filename:
                if out_dir:
                    output_path = os.path.join(out_dir, filename)
                else:
                    output_path = filename
                with open(output_path, "w") as outputbundle:
                    bundle = self.resource_generator.make_bundle()
                    outputbundle.write(bundle.json(indent=2))

        return self.resources, self.resource_generator.make_bundle()

    def generate_patients(self, bundle=True):
        n_patients = math.ceil(float(self.n) / self.n_per_patients)
        patient_generator = PatientGenerator(n=n_patients)
        patients = patient_generator.generate()
        if bundle:
            return patient_generator.make_bundle()
        else:
            return patients

    def update_with_patient_ids(self):
        old_res = self.resources.copy()
        # Step with n per patient
        for index in range(0, len(self.resources), self.n_per_patients):
            patient_resources = self.resources[index: index + self.n_per_patients]
            for resource in patient_resources:
                resource.patient = {
                    "reference": self.patients[int(index / self.n_per_patients)],
                    "type": "Patient"
                }

        print(self.resources)


if __name__ == '__main__':
    # pprint(Patient.schema()["properties"])
    load_dotenv(find_dotenv())
    pg = PatientGenerator(n=100)
    pg.generate()
    bundle = pg.make_bundle()
    server = FhirServer(
        api_address="http://localhost:8080/fhir",
        client_id=os.getenv("CLIENT_ID"),
        client_secret=os.getenv("CLIENT_SECRET"),
        oidc_provider_url=os.getenv("OIDC_PROVIDER_URL")
    )

    response = server.add_bundle(bundle)
    print(response.references)



import random
from typing import List, Tuple, Union

from fhir.resources.observation import Observation, ObservationComponent, ObservationReferenceRange
from fhir_kindling.generators import FhirResourceGenerator
from fhir.resources.patient import Patient
from fhir.resources.codeableconcept import CodeableConcept
from fhir.resources.coding import Coding
from fhir.resources.quantity import Quantity


class ObservationGenerator(FhirResourceGenerator):

    def __init__(self, n: int = None, n_per_patient: int = None, patients: List[Patient] = None,
                 coding_system: str = None, codes: List[str] = None, code_probabilities: List[float] = None,
                 units: Union[List[str], str] = None,
                 value_ranges: Union[List[Tuple[float, float]], Tuple[float, float]] = None):

        super().__init__(n, resource_type=Observation)

        if n and n_per_patient:
            raise ValueError(
                "Both n and n_patients set. Observations can be generated either associated with patients or by "
                "themselves not both")

        if code_probabilities:
            if len(code_probabilities) != len(codes):
                raise ValueError("If probabilities are set they need to match the number of selected codes.\nNumber of"
                                 f"codes: {len(codes)} -- Number of probabilities: {len(code_probabilities)}")

        if n_per_patient and patients is None:
            raise ValueError("No patients given to generate resources for.")

        if units and not value_ranges:
            raise ValueError("Value ranges need to be given when units are given.")
        if value_ranges and not units:
            raise ValueError("Units need to be given if value ranges are given.")

        self.n = n
        self.n_per_patients = n_per_patient
        self.patients = patients
        self.coding_system = coding_system
        self.codes = codes
        self.code_probabilities = code_probabilities
        self.units = units
        self.value_ranges = value_ranges

    def _generate(self):
        if self.n:
            observations = self._generate_observations(self.n)
        elif self.patients:
            observations = self._generate_observations_for_patients()

        return observations

    def _generate_observations(self, n: int) -> List[Observation]:

        observations = []
        # todo add different status
        for i in range(n):
            observation = Observation(
                **{
                    "code": self._generate_code(),
                    "status": "final",
                    "valueQuantity": self._generate_value_quantity()
                }
            )
            observations.append(observation)
        return observations

    def _generate_observations_for_patients(self):
        observations = []

        return observations

    def _generate_code(self) -> CodeableConcept:
        if self.code_probabilities:
            code = random.choices(self.codes, weights=self.code_probabilities, k=1)[0]
        else:
            code = random.choice(self.codes)
        coding = Coding(
            **{
                "system": self.coding_system,
                "code": code
            }
        )

        code = CodeableConcept(
            **{
                "coding": [coding],
                "text": "loinc-code"
            }
        )

        return code

    def _generate_value_quantity(self) -> Quantity:

        if isinstance(self.value_ranges, list):
            index = random.randint(0, len(self.value_ranges))
            value = random.uniform(*self.value_ranges[index])
            unit = self.units[index]
        else:
            value = random.uniform(*self.value_ranges)
            unit = self.units

        quantity = Quantity(
            **{
                "value": value,
                "unit": unit
            }
        )
        return quantity


if __name__ == '__main__':
    code_system = "http://loinc.org"
    observation_codes = ["6598-7", "10839-9", "48425-3", "6597-9", "6598-7", "67151-1", "42757-5", "10839-9", "49563-0"]
    unit = "ng/ml"
    value_range = (0.5, 8.9)
    obs_generator = ObservationGenerator(n=100, coding_system=code_system, codes=observation_codes, units=unit,
                                         value_ranges=value_range)
    res = obs_generator.generate()



import attr
from typing import Union, List


@attr.s
class QueryParameter(object):
    variable: str = attr.ib()
    condition: Union[str, List[str]] = attr.ib()

    @classmethod
    def from_segment(cls, str):
        pass




import math
import pprint
from typing import List, Union
from fhir.resources.domainresource import DomainResource
from fhir.resources.bundle import Bundle, BundleEntry, BundleEntryRequest
from fhir.resources.reference import Reference
import os
import pendulum
from uuid import uuid4

from abc import ABC, abstractmethod

from fhir.resources.resource import Resource


class FhirResourceGenerator:

    def __init__(self, n: int = None, n_per_patient: int = 1, resource_type: DomainResource = None):
        self._patient_references = None
        self.n = n
        self.resource_type = resource_type
        self.resources = None
        self.n_per_patient = n_per_patient

    @property
    def num_patients(self):
        return math.ceil(self.n/self.n_per_patient)

    def generate(self, out_dir: str = None, filename: str = None, generate_ids: bool = False,
                 patient_references: List[Reference] = None) -> List[Resource]:
        self._patient_references = patient_references
        self.resources = self._generate()
        if generate_ids:
            for resource in self.resources:
                resource.id = self.generate_id()

        if self._patient_references:
            self.update_with_patient_ids()

        if out_dir:
            bundle = self.make_bundle()
            if filename:
                path = os.path.join(out_dir, filename)
            else:
                path = os.path.join(out_dir, f"bundle-{pendulum.now().to_date_string()}.json")
            with open(path, "w") as bundle_file:
                bundle_file.write(bundle.json())
        return self.resources

    @abstractmethod
    def _generate(self):
        pass

    def make_bundle(self) -> Bundle:
        entries = self._generate_bundle_entries()
        bundle_data = {
            "type": "transaction",
            "entry": entries
        }
        bundle = Bundle(**bundle_data)
        return bundle

    def _generate_bundle_entries(self):
        entries = []
        for resource in self.resources:
            bundle_entry_dict = {
                "resource": resource,
                "request": BundleEntryRequest(**{"method": "POST", "url": self.resource_type.get_resource_type()})
            }
            entry = BundleEntry(**bundle_entry_dict)
            entries.append(entry)
        return entries

    def update_with_patient_ids(self):
        # Step with n per patient
        for index in range(0, len(self.resources), self.n_per_patient):
            patient_resources = self.resources[index: index + self.n_per_patient]
            for resource in patient_resources:
                resource.patient = {
                    "reference": self._patient_references[int(index / self.n_per_patient)],
                    "type": "Patient"
                }

        print(self.resources)

    def display_schema(self):
        pprint.pprint(self.resource_type.schema())

    @staticmethod
    def generate_id():
        return str(uuid4())



import os
from typing import List, Union

from fhir.resources.bundle import Bundle
from fhir.resources.domainresource import DomainResource
from fhir.resources.molecularsequence import MolecularSequence, MolecularSequenceVariant

from fhir_kindling.generators import FhirResourceGenerator

from dotenv import load_dotenv, find_dotenv


class MolecularSequenceGenerator(FhirResourceGenerator):
    def __init__(self, n: int = None, resources: List[DomainResource] = None,
                 sequence_file: Union[str, List[str]] = None):
        super().__init__(n, resource_type=MolecularSequence)
        self.sequence_file = sequence_file
        if self.sequence_file:
            self.sequences = self._load_sequence_file(self.sequence_file)

    def _generate(self):
        molecular_sequences = []
        for sequence_def in self.sequences:
            mol_seq = self._generate_molecular_sequence(sequence_def[1], sequence_def[2:])
            molecular_sequences.append(mol_seq)
        return molecular_sequences

    def _generate_molecular_sequence(self, sequence: str, variant: List[str]):
        molecular_sequence = MolecularSequence(
            **{
                "coordinateSystem": 0,
                "observedSeq": sequence,
                "variant": self._make_variant_resources(variant)

            }
        )
        return molecular_sequence

    def _make_variant_resources(self, variant: List[str]) -> List[MolecularSequenceVariant]:
        variants = []
        for var in variant:
            variant_resource = MolecularSequenceVariant(
                **{
                    "observedAllele": var
                }
            )
            variants.append(variant_resource)
        return variants

    def _load_sequence_file(self, path: Union[str, List[str]]):

        if isinstance(path, str):
            with open(path, "r") as sf:
                sequences = [line.split() for line in sf.readlines()]

        elif isinstance(path, list):
            sequences = []
            for p in path:
                with open(p, "r") as sf:
                    partial_sequences = [line.split() for line in sf.readlines()]
                    sequences.extend(partial_sequences)

        else:
            raise ValueError("Invalid input arguments needs to be str or list of str")

        self.n = len(sequences)
        return sequences



#!/usr/bin/env python

"""The setup script."""
from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('CHANGELOG.md') as history_file:
    history = history_file.read()

with open("requirements.txt") as requirements_file:
    requirements = requirements_file.readlines()

test_requirements = ['pytest>=3', ]

setup(
    author="Michael Graf",
    author_email='michael.graf@uni-tuebingen.de',
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description="Generates synthetic FHIR resources.",
    entry_points={
        'console_scripts': [
            'fhir_kindling=fhir_kindling.cli.cli:main',
        ],
    },
    long_description=readme + history,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    license="MIT license",
    include_package_data=True,
    keywords='fhir_kindling',
    name='fhir_kindling',
    packages=find_packages(include=['fhir_kindling', 'fhir_kindling.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/migraf/fhir-kindling',
    version='0.2.6',
    zip_safe=False,
)



import json
from typing import Union, List

import click
from fhir.resources.bundle import BundleEntry, Bundle
from pydantic.error_wrappers import ValidationError
from pathlib import Path


def load_bundle(bundle_path: Union[Path, str], validate: bool = True) -> Union[Bundle, dict]:
    """Loads a FHIR bundle stored as a json file from file and returns a bundle object. When validate is selected
    all resource entries in the bundle are individually validated. The full bundle is validated by default.

    Args:
      bundle_path: path to the bundle file
      validate: boolean on whether to validate individual resources

    Returns:
      a bundle object based on the given file

    Raises:
      ValidationError: when FHIR constraints on the bundle are not met

    """
    with open(bundle_path, "r") as f:
        bundle_json = json.load(f)
    if validate:
        bundle = validate_bundle(bundle_json)
        if not bundle:
            raise ValueError("Invalid bundle format")
        return bundle
    return bundle_json


def validate_bundle(bundle_json: dict,
                    show_invalid_entries: bool = True,
                    drop_invalid_entries: bool = False) -> Union[None, Bundle, List[int]]:
    """Validate the all the resources in the bundle and display validation errors.
    If drop

    Args:
      drop_invalid_entries: if True drops all entries in the bundle detected as invalid from the bundle entries
        and attempts tp make a new bundle with only valid entries
      bundle_json: dictionary based on a loaded json bundle
      show_invalid_entries: if True show the entries detected as invalid and the validation errors

    Returns:
        either a Bundle object if the validation succeeded or None if it failed.
    """
    if drop_invalid_entries:
        valid_entries = _validate_bundle_entries(
            bundle_json["entry"],
            show_invalid_entries=show_invalid_entries,
            drop_invalid_entries=drop_invalid_entries)
        bundle_json["entry"] = valid_entries
        try:
            bundle = Bundle(**bundle_json)
            return bundle

        except ValidationError as e:
            click.echo("Bundle could not be validated", err=True)
            if show_invalid_entries:
                click.echo(e, err=True)
            return None

    else:
        errors, error_indices = _validate_bundle_entries(
            bundle_json["entry"],
            show_invalid_entries=show_invalid_entries,
            drop_invalid_entries=drop_invalid_entries
        )

        if errors == 0:
            # Try to validate the full bundle
            try:
                bundle = Bundle(**bundle_json)
                return bundle

            except ValidationError as e:
                click.echo("Bundle could not be validated", err=True)
                if show_invalid_entries:
                    click.echo(e, err=True)
                return None
        else:
            return error_indices


def _validate_bundle_entries(entries: List[dict],
                             show_invalid_entries: bool = True,
                             drop_invalid_entries: bool = False):
    errors = 0
    error_indices = []

    valid_bundle_entries = []

    for index, entry in enumerate(entries):
        try:
            bundle_entry = BundleEntry(**entry)
            valid_bundle_entries.append(bundle_entry)
        # Catch and display validation errors and display them with their index
        except ValidationError as e:
            if show_invalid_entries:
                click.echo(f"Bundle entry with index {index} could not be validated", err=True)
                click.echo(f"\n {e} \n", err=True)
                click.echo(json.dumps(entry, indent=2), err=True)
            errors += 1
            error_indices.append(index)

    # drop invalid is set return only the validated entries
    if drop_invalid_entries:
        return valid_bundle_entries

    return errors, error_indices
